#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/vector.hpp>
#include <boost/numeric/ublas/assignment.hpp>
#include <boost/numeric/ublas/io.hpp>
#include <iomanip>

using namespace boost::numeric::ublas;

//TODO: USE UBLAS OR BOOST TEST FRAMEWORK

template <class AE>
typename AE::value_type mean_square(const matrix_expression<AE> &me) {
    typename AE::value_type s(0);
    typename AE::size_type i, j;
    for (i=0; i!= me().size1(); i++) {
        for (j=0; j!= me().size2(); j++) {
            s+=std::fabs(me()(i,j));
        }
    }
    return s/me().size1()*me().size2();
}

template <class AE>
typename AE::value_type mean_square(const vector_expression<AE> &ve) {
    // We could have used norm2 here, but ublas' ABS does not support unsigned types.
    typename AE::value_type s(0);
    typename AE::size_type i;
    for (i=0; i!= ve().size(); i++) {
            s+=std::fabs(ve()(i));
        }
    return s/ve().size();
}

template <class T> struct eps {} ;

template <> struct eps<double> { static const double value = 1.0e-12; };

template <> struct eps<float> { static const float value = 1.0e-7; };

template <> struct eps<long> {  static const long value = 1; }; //Don't expect to do arithmetics with long

template <> struct eps<int> {  static const int value = 1; }; //Don't expect to do arithmetics with int

template <class value_type>
void test_fixed_vector() {

    value_type prec = eps<value_type>::value;

    fixed_vector<value_type, 3> v1, v2, v3;
    fixed_vector<value_type, 4> v4, v5, v6, v7;

    // fixed vector size. There is a chance that only EBO compilers (most are) will satisfy this:
    assert(sizeof(fixed_vector<value_type, 3>) == sizeof(value_type)*3);

    // Assignment
    v1 <<= 1, 2.76, 3;
    v2 <<= 1, 2.76, 3;
    v3 = v1; // TODO: THIS SHOULD BE AS FAST AS ASSIGNING TYPES LIKE struct {double x,y,z;};
    assert(mean_square(v2-v3)<prec);

    // Assignment to differnet size
    bool catched=  false;
    try {
        v4=v3;
    } catch (...) {
        std::cerr << "CHECKED FAILED IS WHAT WE WANT HERE" << std::endl;
        catched = true;
    }
    assert(catched);

    // A non-trivial expression
    v4 <<= 5.0,     3.0,    12.66,  3.54;
    v5 <<= 122.0,   45.0,   154.33, 6.77;
    v6 = 2.0*v4+3.0*v5-inner_prod(v4,v5)*v5;
    std::cout << std::setprecision(20) << v6 << std::endl;
    v7 <<= -331803.59919999999693,-122384.26200000000244,-419718.88298800005578,-18405.854972000000998;
    assert(mean_square(v6-v7)<mean_square(v7)*prec); //This assertion is oversimplified


    // Mixing
    vector<double> b(3);
    b<<=2.0, 4.33, 5.77;
    v2 <<=2.0, 4.33, 5.77;
    v1 = b;
    assert(mean_square(v2-v1)<prec);
    v2 = 2.0*v1 + b;
    v3 <<= 6, 12.990000000000000213, 17.309999999999998721;
    std::cout << std::setprecision(20) << v2 << std::endl;
    assert(mean_square(v3-v2)<mean_square(v3)*prec); //This assertion is oversimplified


    //TODO:  need to add some more blas2 and blas3 tests
}

template <class value_type>
void test_fixed_matrix() {

    value_type prec = eps<value_type>::value;

    fixed_matrix<value_type, 3, 3> m1, m2, m3;
    fixed_matrix<value_type, 2, 3> m4, m5, m6, m7;

    // fixed matrix size. There is a chance that only EBO compilers (most are) will satisfy this:
    assert(sizeof(fixed_matrix<value_type, 3, 3>) == sizeof(value_type)*3*3);

    // Assignment
    m1 <<= 1,   2.76, 3,
           2,   5.55, 7,
           8.1, 4.32, 8.54;

    m2 <<= 1,   2.76, 3,
           2,   5.55, 7,
           8.1, 4.32, 8.54;
    m3 = m1; // TODO: THIS SHOULD BE AS FAST AS ASSIGNING TYPES LIKE struct {double a,b,c,d,e,f,g,h,i;};
    std::cout << m3 << std::endl;
    assert(mean_square(m2-m3)<prec);

    // Assignment to differnet size
    bool catched=  false;
    try {
        m4=m3;
    } catch (...) {
        std::cerr << "CHECKED FAILED IS WHAT WE WANT HERE" << std::endl;
        catched = true;
    }
    assert(catched);

    // A non-trivial expression
    m4 <<= 5.0,     3.0,    12.66,
           3.54,    2.0,    7.84;
    m5 <<= 122.0,   45.0,   154.33,
           6.77,    5.66,   9.62;

    std::cout << m4 << std::endl << m5 << std::endl;
    m6 = 2.0*m4-3.0*m5;
    std::cout << std::setprecision(20) << m6 << std::endl;
    m7 <<=  -356,                   -129,                   -437.67000000000001592
            ,-13.22999999999999865, -12.980000000000000426, -13.179999999999999716;
    assert(mean_square(m6-m7)<mean_square(m7)*prec); //This assertion is oversimplified

    // Mixing
    matrix<double> b(3,3);
    b <<=   2.0,    4.33, 5.77,
            3.22,   6.44, 7.51,
            9.12,   4.66, 4.98;

    m2 <<=  2.0,    4.33, 5.77,
            3.22,   6.44, 7.51,
            9.12,   4.66, 4.98;
    m1 = b;
    assert(mean_square(m2-m1)<prec);

    m1 <<= 1,   2.76, 3,
           2,   5.55, 7,
           8.1, 4.32, 8.54;

    m2 = 2.0*m1 + b;
    m3 <<=  4,                      9.8499999999999996447,  11.769999999999999574,
            7.2200000000000006395,  17.539999999999999147,  21.50999999999999801,
            25.320000000000000284,  13.300000000000000711,  22.059999999999998721;

    std::cout << std::setprecision(20) << m2 << std::endl;
    assert(mean_square(m3-m2)<mean_square(m3)*prec); //This assertion is oversimplified

    //TODO: need to add some more blas2 and blas3 tests. Need to add column major tests
}

int main() {

    test_fixed_vector<double>();
    test_fixed_vector<float>();
    test_fixed_vector<long>();
    test_fixed_vector<int>();

    test_fixed_matrix<double>();
    test_fixed_matrix<float>();
    test_fixed_matrix<long>();
    test_fixed_matrix<int>();

    return 0;
}

